/* Copyright (C) 2002 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: conference.c,v $
   Revision 1.5  2002/08/18 16:42:06  wb6ymh
   Automatically exit lurking mode when a lurker talks.
   Corrected a bug that allowed two stations to talk at once when they doubled.

   Revision 1.4  2002/08/15 05:36:51  wb6ymh
   Fix bug that allows timed out talker to keep transmit token forever.
   Added version command.
   Added "?" command (alias for help command).

   Revision 1.3  2002/08/12 17:08:35  wb6ymh
   Removed some left over debug info from log.
   Removed \r from end of commands before logging them.
   Added abilty to record and playback audio from files.
   Added .help and .test commands.

   Revision 1.2  2002/08/10 20:52:06  wb6ymh
   disable dumping of avl trees

   Revision 1.1.1.1  2002/08/10 20:33:41  wb6ymh
   initial import

*/

#include "common.h"

#ifndef _WIN32
   // FreeBSD, Linux, etc..
   #include <stdio.h>
   #include <stdlib.h>
   #include <sys/types.h>
   #ifdef TIME_WITH_SYS_TIME
      #include <sys/time.h>
      #include <time.h>
   #else
      #ifdef HAVE_SYS_TIME_H
         #include <sys/time.h>
      #else
         #include <time.h>
      #endif
   #endif
   #include <unistd.h>
   #include <sys/socket.h>
   #include <netinet/in.h>
   #include <arpa/inet.h>
   #include <errno.h>
   #include <string.h>
#ifdef HAVE_FCNTL_H
   #include <fcntl.h>
#endif
   #include <netdb.h>
   #include <ctype.h>
   #ifdef HAVE_SYS_TIMEB_H
      #include <sys/timeb.h>
   #endif
#else
   // Windoze
   #include <stdio.h>
   #include <time.h>
   #include <io.h>
   #include <winsock2.h>
   #include <sys\timeb.h>
#endif

#include "avl.h"
#include "main.h"
#include "dirclient.h"
#include "rtp.h"
#include "configvars.h"
#include "conference.h"
#include "ilink.h"

#define  CONF_BUF_SIZE     1024

#define  DUMP_TREES
#undef   DUMP_TREES
typedef enum {
   EVENT_INIT = 1,
   EVENT_RTP_RX,
   EVENT_RTP_TO,
   EVENT_RTCP_RX,
   EVENT_RTCP_TO,
} EventType;

typedef struct ConfClient_TAG {
   IPAdrUnion AdrRTP;
   IPAdrUnion AdrRTCP;
   int      SN;
   int (*State)(ClientInfo *p,struct ConfClient_TAG *pCC,EventType Event);
   time_t   LastHeard;
   char  *CallPlus;
   char  *Callsign;
   FILE  *fp;
   struct timeval TimeFirstTx;
   struct timeval TimeLastTx;
   struct timeval LastAudioIn;
   int   Timeout;
   void  *p;         // various uses

   int   bInConf:1;
   int   bLurking:1;
   int   bTalking:1;
   int   bAdmin:1;
} ConfClient;

struct avl_table *ConfTree;   // sorted by IP address

u_long ClientTalking = 0;
struct timeval LastAudioIn;
char *OurSDES = NULL;
int SDESLen;

const char *CallSignString = "CALLSIGN";

char StatusMsg[512];
int StatusMsgLen = 0;

time_t NextLoginTime;
time_t NextStationListTime;

int ConferenceClients;
int ClientConnects;

int CClientCompare(const void *avl_a,const void *avl_b,void *avl_param);
ConfClient *CreateNewConfClient(void);
void DeleteCClient(ConfClient *pCC);
int CheckRTCP(rtcp_common_t *p,int len);
char *GetCall(rtcp_t *p,int len);
char *GetLocalIPAdr(void);
int GenSDES(char *Temp);
int GenBye(char *Temp,char *Reason);
void SendStationList(ClientInfo *p);
void SetTimeoutRTCP(ClientInfo *p);
void ConferenceCmd(ClientInfo *p,ConfClient *pCC,char *cmd);
void DumpTree();
int TimeLapse(struct timeval *p);
void CmdHelp(ClientInfo *p,ConfClient *pCC1,const char *Arg);

int ConferenceInit()
{
   IPAdrUnion MyAdr;
   ClientInfo *pRTP;
   ClientInfo *pRTCP;
   time_t Time = time(NULL);
   char Temp[512];

   ConfTree = avl_create(CClientCompare,NULL,NULL);

   if(ConfTree == NULL) {
      LOG_ERROR(("ConferenceInit(): avl_create failed.\n"));
      return FALSE;
   }

   pRTP = CreateNewClient();
   if(pRTP == NULL) {
      LOG_ERROR(("ConferenceInit(): CreateNewClient failed for RTP.\n"));
      return FALSE;
   }

   pRTP->BufSize = CONF_BUF_SIZE;
   pRTP->Buf = malloc(pRTP->BufSize);
   
   pRTP->Socket = socket(AF_INET,SOCK_DGRAM,0);
   if(pRTP->Socket == SOCKET_ERROR) {
      LOG_ERROR(("ConferenceInit(): Socket() failed for RTP\n"));
      return FALSE;
   }

   MyAdr.i.sin_family = AF_INET;
   MyAdr.ADDR = INADDR_ANY;
   pRTP->HisAdr.PORT = MyAdr.PORT = htons(ILINK_RTP_PORT);

   if(bind(pRTP->Socket,&MyAdr.s,sizeof(MyAdr)) == SOCKET_ERROR) {
      LOG_ERROR(("ConferenceInit(): bind() failed for RTP.\n"));
      DeleteClient(pRTP);
      return FALSE;
   }
   SET_WAIT4_RD(pRTP);
   pRTP->State = RTP_Handler;

   pRTCP = CreateNewClient();
   if(pRTCP == NULL) {
      LOG_ERROR(("ConferenceInit(): CreateNewClient failed for RTP.\n"));
      return FALSE;
   }

   pRTCP->BufSize = CONF_BUF_SIZE;
   pRTCP->Buf = malloc(pRTCP->BufSize);

   pRTCP->Socket = socket(AF_INET,SOCK_DGRAM,0);

   if(pRTCP->Socket == SOCKET_ERROR) {
      LOG_ERROR(("ConferenceInit(): Socket() failed for RTCP.\n"));
      return FALSE;
   }

   pRTCP->HisAdr.PORT = MyAdr.PORT = htons(ILINK_RTCP_PORT);

   if(bind(pRTCP->Socket,&MyAdr.s,sizeof(MyAdr)) == SOCKET_ERROR) {
      LOG_ERROR(("ConferenceInit(): bind() failed for RTCP.\n"));
      DeleteClient(pRTP);
      DeleteClient(pRTCP);
      return FALSE;
   }

   SET_WAIT4_RD(pRTCP);
   pRTCP->State = RTCP_Handler;
   
   avl_insert(ClientTree,pRTP);
   avl_insert(ClientTree,pRTCP);

   SDESLen = GenSDES(Temp);
   OurSDES = (char *) malloc(SDESLen);
   if(OurSDES == NULL) {
      return FALSE;
   }
   memcpy(OurSDES,Temp,SDESLen);

   
   if(LoginInterval > 0) {
      NextLoginTime = Time + LoginInterval;
   }
   else {
      NextLoginTime = 0;
   }

   if(StationListInterval > 0) {
      NextStationListTime = Time + StationListInterval;
   }
   else {
      NextStationListTime = 0;
   }

   if(StationListInterval > 0 && LoginInterval > 0) {
      ServerRequest(SERV_REQ_LOGIN_AND_LIST);
   }
   else if(LoginInterval > 0) {
      ServerRequest(SERV_REQ_LOGIN);
   }
   SetTimeoutRTCP(pRTCP);

   return TRUE;
}

int RTP_Rx(ClientInfo *p)
{
   int RxLen;
   ConfClient  Lookup;
   ConfClient *pCC = NULL;
   ConfClient *pNextCC = NULL;
   int Adrlen = sizeof(Lookup.AdrRTCP);
   struct avl_traverser avl_trans;
   int bSendStationList = FALSE;
   char *cp;
   char *cp1;

   RxLen = recvfrom(p->Socket,p->Buf,p->BufSize-1,0,&Lookup.AdrRTCP.s,&Adrlen);
   p->Buf[RxLen] = 0;
   p->Count = RxLen;

   if(RxLen == SOCKET_ERROR) {
      DPRINTF(("RTP_Handler(): recvfrom returned %u.\n",ERROR_CODE));
   }
   else if((pCC = avl_find(ConfTree,&Lookup)) != NULL) {
      if(pCC->State != NULL) {
         pCC->State(p,pCC,EVENT_RTP_RX);
      }
      else if(p->Buf[0] == ILINK_DATA_PACKET) {
      // Data packet.  We only want to forward message line traffic, 
      // not the general station info that we get when a new station joins
      // the conference.  The general form of message line traffic is
      // "callsign>data ..."  Check for "callsign>" at the begining of
      // the line.
         cp = &p->Buf[strlen(NDATA)+1];   // Point to first character of data
         if((cp1 = strchr(cp,'>')) != NULL) {
            *cp1 = 0;
            if(strncmp(cp,pCC->Callsign,strlen(cp)) == 0) {
            // Yup, it's a message line
               *cp1 = '>';
               if(cp1[1] == '.') {
                  if((cp = strchr(cp,'\r')) != NULL) {
                     *cp = 0;
                  }
                  LOG_NORM(("%s said \"%s\".\n",pCC->Callsign,&cp1[1]));
                  ConferenceCmd(p,pCC,&cp1[2]);
               }
               else {
                  pNextCC = (ConfClient *) avl_t_first(&avl_trans,ConfTree);
               }
            }
         }
      }
      else {
      // Audio packet
         pCC->LastAudioIn = TimeNow;
         pCC->bTalking = TRUE;
         if(ClientTalking == 0) {
         // A new station is talking
            SetTimeout(p,ConfAudioTimeout);
            ClientTalking = pCC->AdrRTP.ADDR;
            bSendStationList = TRUE;
            pCC->bLurking = FALSE;
         }

         if(pCC->AdrRTP.ADDR == ClientTalking) {
            LastAudioIn = TimeNow;
            pNextCC = (ConfClient *) avl_t_first(&avl_trans,ConfTree);
         }
      }

      while((pCC = pNextCC) != NULL) {
         pNextCC = (ConfClient *) avl_t_next(&avl_trans);
         if((TimeNow.tv_sec - pCC->LastHeard) > ConfMemberTimeout) {
         // timeout
            LOG_NORM(("Dropping %s from conference, RTP time out.\n",
                      pCC->Callsign));
            if(pCC->State != NULL) {
               pCC->State(p,pCC,EVENT_RTCP_TO);
            }
            DeleteCClient(pCC);
         }
         else if(pCC->bInConf) {
            if(p->Buf[0] == ILINK_DATA_PACKET || 
               pCC->AdrRTP.ADDR != ClientTalking)
            {
               sendto(p->Socket,p->Buf,RxLen,0,&pCC->AdrRTP.s,
                      sizeof(pCC->AdrRTP));
            }
         }
      }
   }
   else {
      LOG_NORM(("RTP_Rx: Failed to find client %s\n",
                inet_ntoa(Lookup.AdrRTCP.i.sin_addr)));
   }

   return bSendStationList;
}

int RTP_Handler(ClientInfo *p)
{
   int bSendStationList = FALSE;
   int bClientsTalking = FALSE;
   int bConfTalkerFound = FALSE;
   int AudioDeltaT;
   struct avl_traverser avl_trans;
   ConfClient *pCC = avl_t_first(&avl_trans,ConfTree);
   
   GetTimeNow();

// Check RTP timeouts
   while(pCC != NULL) {
      if(pCC->bTalking) {
         AudioDeltaT = TimeLapse(&pCC->LastAudioIn);

         if(AudioDeltaT > ConfAudioTimeout) {
         // This guy has timed out
            pCC->bTalking = FALSE;

            if(pCC->State != NULL) {
               pCC->State(p,pCC,EVENT_RTP_TO);
            }
         }
         else {
         // This guy is still talking
            bClientsTalking = TRUE;
            if(pCC->AdrRTP.ADDR == ClientTalking) {
               bConfTalkerFound = TRUE;
            }
         }
      }
      pCC = (ConfClient *) avl_t_next(&avl_trans);
   }

   if(!bConfTalkerFound) {
   // The conference is free now
      ClientTalking = 0;
      bSendStationList = TRUE;
   }

   if(bClientsTalking) {
   // Set a timeout
      SetTimeout(p,ConfAudioTimeout);
   }
   else {
   // No one's talking no need for a timeout
      p->bTimeWait = FALSE;
   }
   
   if(p->bReadReady) {
      bSendStationList |= RTP_Rx(p);
   }

   if(bSendStationList) {
      SendStationList(p);
   }

   return FALSE;
}


int RTCP_Rx(ClientInfo *p)
{
   int RxLen;
   int Ret = FALSE;
   ConfClient  Lookup;
   char *CallSign;
   char *end;
   ConfClient *pCC;
   int TxLen;
   int Adrlen = sizeof(Lookup.AdrRTCP);
   char *cp;
   union {
      rtcp_common_t *p;
      rtcp_t *pRTCP;
      char *cp;
   } u;

   union {
      rtcp_sdes_item_t *p;
      char *cp;
   } Item;
   
   u.cp = p->Buf;

   RxLen = recvfrom(p->Socket,p->Buf,p->BufSize,0,&Lookup.AdrRTCP.s,&Adrlen);
   p->Count = RxLen;

   if(RxLen == SOCKET_ERROR) {
      DPRINTF(("RTCP_Handler(): recvfrom returned %u.\n",ERROR_CODE));
      return FALSE;
   }

   if(!CheckRTCP(u.p,RxLen)) {
   // Bogus packet, ignore it
      DPRINTF(("RTCP_Handler(): ignoring bogus packet.\n"));
      return FALSE;
   }

   if((pCC = avl_find(ConfTree,&Lookup)) == NULL) {
   // New conference client
      if((CallSign = GetCall(u.pRTCP,RxLen)) != NULL) {
         if((cp = strchr(CallSign,' ')) != NULL) {
         // Remove the name from the callsign/name string
            *cp = 0;
         }

         if(ConferenceClients < MaxConferenceClients) {
            ConferenceClients++;
            pCC = CreateNewConfClient();
            pCC->AdrRTP = Lookup.AdrRTCP;
            pCC->AdrRTCP = Lookup.AdrRTCP;
            pCC->AdrRTP.PORT = htons(ILINK_RTP_PORT);
            pCC->AdrRTCP.PORT = htons(ILINK_RTCP_PORT);
            pCC->Callsign = strdup(CallSign);
            if(cp != NULL) {
            // Restore the name to the end of the callsign
               *cp = ' ';
            }
            pCC->CallPlus = CallSign;
            LOG_NORM(("%s logged into conference.\n",pCC->Callsign));
#ifdef DUMP_TREES
            DumpTree();
            avl_insert(ConfTree,pCC);
            LOG_NORM(("After insert:\n"));
            DumpTree();
#else
            avl_insert(ConfTree,pCC);
#endif
            SendStationList(p);
         }
         else {
         // Sorry Jack, we're full
            LOG_NORM(("%s failed to login when conference was full.\n",
                      CallSign));
            TxLen = GenBye(p->Buf,"Sorry the conference is full");
            Lookup.AdrRTCP.PORT = htons(ILINK_RTCP_PORT);
            sendto(p->Socket,p->Buf,TxLen,0,&Lookup.AdrRTCP.s,
                   sizeof(Lookup.AdrRTCP));
            free(CallSign);
         }
      }
   }
   
   if(pCC != NULL) {
      pCC->LastHeard = time(NULL);

      if(pCC->State != NULL) {
         pCC->State(p,pCC,EVENT_RTCP_RX);
      }

      Item.p = NULL;
      end = u.cp + RxLen;

      while(u.cp < end) {
         D2PRINTF(("RTCP_Handler(): processing pt %d.\n",u.p->pt));
         switch(u.p->pt) {
            case RTCP_SR:
               break;

            case RTCP_RR:
               break;

            case RTCP_SDES:
            // Answer with our SDES
               sendto(p->Socket,OurSDES,SDESLen,0,&pCC->AdrRTCP.s,
                      sizeof(pCC->AdrRTCP));
               break;

            case RTCP_BYE:
               LOG_NORM(("%s leaving conference.\n",pCC->Callsign));
               DeleteCClient(pCC);
            // Update the station list
               SendStationList(p);
               return Ret;

            case RTCP_APP:
               break;
         }
         u.cp = u.cp + sizeof(rtcp_common_t) + ntohs(u.p->length) * 4;
      }
   }

   return Ret;
}

void SetTimeoutRTCP(ClientInfo *p)
{
   if(NextLoginTime != 0 && NextLoginTime < NextStationListTime) {
      SetTimeout(p,(NextLoginTime - TimeNow.tv_sec) * 1000);
   }
   else if(NextStationListTime != 0) {
      SetTimeout(p,(NextStationListTime - TimeNow.tv_sec) * 1000);
   }
   else {
      p->bTimeWait = FALSE;
   }
}

int RTCP_Handler(ClientInfo *p)
{
   int Ret = FALSE;

   if(p->bReadReady) {
      Ret |= RTCP_Rx(p);
   }

   if(p->bTimeOut) {
   // refresh our entry in the directory server
      if(NextLoginTime <= TimeNow.tv_sec) {
      // Time to login again
         NextLoginTime = TimeNow.tv_sec + LoginInterval;
         if(NextStationListTime != 0 && NextStationListTime <= TimeNow.tv_sec) {
         // Also time to refresh the station list
            NextStationListTime = TimeNow.tv_sec + StationListInterval;
            DPRINTF(("RTCP_Handler(): refreshing login and station list.\n"));
            ServerRequest(SERV_REQ_LOGIN_AND_LIST);
         }
         else {
            DPRINTF(("RTCP_Handler(): refreshing login.\n"));
            ServerRequest(SERV_REQ_LOGIN);
         }
      }
      else if(NextStationListTime != 0 && NextStationListTime <= TimeNow.tv_sec) {
      // Time to refresh our station list
         NextStationListTime = TimeNow.tv_sec + StationListInterval;
         DPRINTF(("RTCP_Handler(): refreshing station list.\n"));
         ServerRequest(SERV_REQ_STATION_LIST);
      }
      
      SetTimeoutRTCP(p);
   }

   return Ret;
}


// Client compare by IP address
int CClientCompare(const void *avl_a,const void *avl_b,void *avl_param)
{
   ConfClient *p_a = (ConfClient *) avl_a;
   ConfClient *p_b = (ConfClient *) avl_b;

   if(p_a->AdrRTCP.ADDR == p_b->AdrRTCP.ADDR) {
      return 0;
   }
   if(p_a->AdrRTCP.ADDR > p_b->AdrRTCP.ADDR) {
      return 1;
   }
   else {
      return -1;
   }
}


ConfClient *CreateNewConfClient()
{
   ConfClient *pCC = (ConfClient *) malloc(sizeof(ConfClient));

   if(pCC != NULL) {
      memset(pCC,0,sizeof(ConfClient));
      pCC->SN = ++ClientConnects;
      pCC->bInConf = TRUE;
   }

   return pCC;
}

void DeleteCClient(ConfClient *pCC)
{
   size_t Nodes = ConfTree->avl_count;

#ifdef DUMP_TREES
   LOG_NORM(("DeleteCClient: deleting pCC->SN=%d, pCC=0x%x\n",pCC->SN,pCC));
   DumpTree();
#endif
   
   if(avl_delete(ConfTree,pCC) == NULL) {
      LOG_ERROR(("DeleteCClient: avl_delete() failed to find client.\n"));
      DumpTree();
   }
   else {
#ifdef DUMP_TREES
      LOG_NORM(("DeleteCClient: After successful delete\n"));
      DumpTree();
#endif

      if((Nodes - 1) != ConfTree->avl_count) {
         LOG_NORM(("DeleteCClient: %d nodes before delete, %d after\n",Nodes,
                   ConfTree->avl_count));
      }

      ConferenceClients--;

      if(pCC->fp != NULL) {
         fclose(pCC->fp);
      }

      if(pCC->CallPlus != NULL) {
         free(pCC->CallPlus);
         pCC->CallPlus = NULL;
      }

      if(pCC->Callsign != NULL) {
         free(pCC->Callsign);
         pCC->Callsign = NULL;
      }
      free(pCC);
   }
}

// Sanity check RTCP packet
int CheckRTCP(rtcp_common_t *p,int len)
{
   char *end;
   union {
      rtcp_common_t *p;
      char *cp;
   } u;
    
   u.p = p;
   end = u.cp + len;

// Sanity check: Version correct, no padding in first packet,
// first item is either RTCP_SR or RTCP_RR

   while(u.cp < end) {
      if(u.p == p && (p->p || (p->pt != RTCP_SR && p->pt != RTCP_RR)) ){
      // Invalid RTCP packet
         return FALSE;
      }

      if(u.p->version != ILINK_RTP_VERSION) {
      // Invalid RTCP packet
         return FALSE;
      }
      u.cp = u.cp + sizeof(rtcp_common_t) + ntohs(u.p->length) * 4;
   }

   return u.cp == end;
}
      
char *GetCall(rtcp_t *p,int len)
{
   char *end;
   char *Callsign = NULL;
   union {
      rtcp_common_t *p;
      rtcp_t *pRTCP;
      char *cp;
   } u;

   union {
      rtcp_sdes_item_t *p;
      char *cp;
   } Item;
   int  bFoundCallsign = FALSE;
    
   Item.p = NULL;
   u.pRTCP = p;
   end = u.cp + len;

   while(u.cp < end) {
      if(u.p->pt == RTCP_SDES) {
      // SDES packet
         Item.p = u.pRTCP->r.sdes.item;
         break;
      }
      u.cp = u.cp + sizeof(rtcp_common_t) + ntohs(u.p->length) * 4;
   }

   while(Item.p != NULL && Item.p->type != RTCP_SDES_END) {
      switch(Item.p->type) {
         case RTCP_SDES_CNAME:
            if(Item.p->length == 8 && strncmp(Item.p->data,"CALLSIGN",8) == 0) {
               bFoundCallsign = TRUE;
            }
            break;
         
         case RTCP_SDES_NAME:
            if(Item.p->length > 0) {
               if(Callsign != NULL) {
               // Two NAME fields !
                  free(Callsign);
               }

               Callsign = malloc(Item.p->length+1);

               if(Callsign != NULL) {
                  memcpy(Callsign,Item.p->data,Item.p->length);
                  Callsign[Item.p->length] = 0;
               }
            }
            break;
      }
      Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;
   }

   if(!bFoundCallsign) {
      if(Callsign != NULL) {
         free(Callsign);
         Callsign = NULL;
      }
   }

   return Callsign;
}

void SendStationList(ClientInfo *p)
{
   ConfClient *pCC;
   struct avl_traverser avl_trans;
   char *pWr = StatusMsg;
   char *pEnd = &StatusMsg[sizeof(StatusMsg)-1];

   *pWr++ = ILINK_DATA_PACKET;

   pWr += sprintf(pWr,NDATA "CONF %s%s\r\r",ConferenceID,
                  ClientTalking == 0 ? "  <SB>" : "");

   pCC = (ConfClient *) avl_t_first(&avl_trans,ConfTree);
   while(pCC != NULL && pWr < pEnd) {
      if(pCC->AdrRTP.ADDR == ClientTalking) {
         pWr += snprintf(pWr,pEnd-pWr,"%s <\r",pCC->CallPlus);
      }
      else if(pCC->bInConf && !pCC->bLurking) {
         pWr += snprintf(pWr,pEnd-pWr,"%s\r",pCC->CallPlus);
      }
      pCC = (ConfClient *) avl_t_next(&avl_trans);
   }

   if(pWr < pEnd) {
      *pWr++ = 0;
      StatusMsgLen = pWr - StatusMsg;
   }
   else {
   // Overflow !
      LOG_ERROR(("GenStationList(): buffer overflow.\n"));
      StatusMsgLen = 0;
   }

   if(StatusMsgLen > 0) {
      pCC = (ConfClient *) avl_t_first(&avl_trans,ConfTree);
   }

   while(pCC != NULL) {
      if(pCC->bInConf) {
         sendto(p->Socket,StatusMsg,StatusMsgLen,0,&pCC->AdrRTP.s,
                sizeof(pCC->AdrRTP));
      }
      pCC = (ConfClient *) avl_t_next(&avl_trans);
   }
}

int GenSDES(char *Temp)
{
   int Len;
   char PadCount;
   char *pFirstData;
   struct tm *tm = localtime(&TimeNow.tv_sec);

   union {
      char *cp;
      rtcp_t *p;
   } u;

   union {
      char *cp;
      rtcp_sdes_item_t *p;
   } Item;

   u.cp = Temp;

   u.p->common.version = ILINK_RTP_VERSION;
   u.p->common.p = 0;
   u.p->common.count = 0;
   u.p->common.pt = RTCP_RR;
   u.p->common.length = htons(1);
   u.p->r.rr.ssrc = 0;

   u.cp += sizeof(u.p->common) + sizeof(u.p->r.rr.ssrc);

   u.p->common.version = ILINK_RTP_VERSION;
   u.p->common.p = 1;
   u.p->common.count = 1;
   u.p->common.pt = RTCP_SDES;

   u.p->r.sdes.src = 0;
   pFirstData = (char *) &u.p->r.sdes.src;

   Item.p = &u.p->r.sdes.item[0];
   Item.p->type = RTCP_SDES_CNAME;
   Item.p->length = (u_int8) strlen(CallSignString);
   memcpy(Item.p->data,CallSignString,Item.p->length);
   
   Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;
   Item.p->type = RTCP_SDES_NAME;
   Item.p->length = (u_int8) strlen(ConferenceCall);
   memcpy(Item.p->data,ConferenceCall,Item.p->length);

   Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;
   Item.p->type = RTCP_SDES_EMAIL;
   Item.p->length = (u_int8) strlen(CallSignString);
   memcpy(Item.p->data,CallSignString,Item.p->length);

   Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;
   Item.p->type = RTCP_SDES_PHONE;
   sprintf(Item.p->data,"%02d:%02d",tm->tm_hour,tm->tm_min);
   Item.p->length = (u_int8) strlen(Item.p->data);
   Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;

   Item.p->type = RTCP_SDES_END ;
   Item.p->length = 0;
   Item.cp = Item.cp + sizeof(rtcp_sdes_item_t) + Item.p->length - 1;

   Len = Item.cp - pFirstData;
   PadCount = Len % 4;

   if(PadCount > 0) {
   // Padding is needed
      u.p->common.p = 1;
      if(PadCount > 1) {
         memset(Item.cp,0,PadCount-1);
      }
      Item.cp[PadCount-1] = PadCount;
      Item.cp += PadCount;
   }
   else {
   // No padding needed
      u.p->common.p = 0;
   }
   u.p->common.length = htons((u_short) ((Item.cp - pFirstData)/4));
   return Item.cp - Temp;
}

int GenBye(char *Temp,char *Reason)
{
   int Len;
   char PadCount;
   char *pFirstData;
   char *cp;

   union {
      char *cp;
      rtcp_t *p;
   } u;

   u.cp = Temp;

   u.p->common.version = ILINK_RTP_VERSION;
   u.p->common.p = 0;
   u.p->common.count = 0;
   u.p->common.pt = RTCP_RR;
   u.p->common.length = htons(1);
   u.p->r.rr.ssrc = 0;

   u.cp += sizeof(u.p->common) + sizeof(u.p->r.rr.ssrc);

   u.p->common.version = ILINK_RTP_VERSION;
   u.p->common.count = 1;
   u.p->common.pt = RTCP_BYE;

   u.p->r.bye.src[0] = 0;
   pFirstData = cp = (char *) &u.p->r.bye.src + sizeof(u.p->r.bye.src);

   strcpy(cp,Reason);

   Len = strlen(Reason) + 1;
   cp += Len;
   PadCount = Len % 4;

   if(PadCount > 0) {
   // Padding is needed
      u.p->common.p = 1;
      if(PadCount > 1) {
         memset(cp,0,PadCount-1);
      }
      cp[PadCount-1] = PadCount;
      cp += PadCount;
   }
   else {
   // No padding needed
      u.p->common.p = 0;
   }
   u.p->common.length = htons((u_short) ((cp - pFirstData)/4));
   return cp - Temp;
}



void CmdLurk(ClientInfo *p,ConfClient *pCC,const char *Arg)
{
   if(!pCC->bLurking) {
      pCC->bLurking = TRUE;
      SendStationList(p);
   }
}

void CmdDelurk(ClientInfo *p,ConfClient *pCC,const char *Arg)
{
   if(pCC->bLurking) {
      pCC->bLurking = FALSE;
      SendStationList(p);
   }
}

void CmdLurkers(ClientInfo *p,ConfClient *pCC1,const char *Arg)
{
   ConfClient *pCC;
   struct avl_traverser avl_trans;
   int Lurkers = 0;


   pCC = (ConfClient *) avl_t_first(&avl_trans,ConfTree);
   while(pCC != NULL) {
      if(pCC->bInConf && pCC->bLurking) {
         if(Lurkers == 0) {
            BufPrintf(p,"Lurkers:\r");
         }
         BufPrintf(p,"%s\r",pCC->CallPlus);
         Lurkers++;
      }
      pCC = (ConfClient *) avl_t_next(&avl_trans);
   }

   if(Lurkers == 0) {
      BufPrintf(p,"I see no lurkers here, try down below\r");
   }
}


void CmdUpTime(ClientInfo *p,ConfClient *pCC1,const char *Arg)
{
   int Days, Hours, Minutes;
   time_t Seconds = time(NULL) - BootTime;

   Days = Seconds / (60 * 60 * 24);
   Seconds %= 60 * 60 * 24;
   Hours = Seconds / (60 * 60);
   Seconds %= 60 * 60;
   Minutes = Seconds / 60;

   BufPrintf(p,"Uptime: %d days, %d hours and %d minutes\r",Days,Hours,Minutes);
}

void CmdVersion(ClientInfo *p,ConfClient *pCC1,const char *Arg)
{
   BufPrintf(p,PACKAGE " V " VERSION " on " OUR_HOST);
}

void PlayBackComplete(ConfClient *pCC)
{
   ClientInfo *pClient;

   if(pCC->fp != NULL) {
      fclose(pCC->fp);
      pCC->fp = NULL;
   }
   
   if(pCC->p != NULL) {
      pClient = (ClientInfo *) pCC->p;
      pClient->Socket = INVALID_SOCKET;
      DeleteClient(pClient);
      pCC->p = NULL;
   }
   pCC->bInConf = TRUE;
   pCC->State = NULL;
}

int SendAudioFromFile(ClientInfo *p,ConfClient *pCC)
{
   int Len;
   int Ret = FALSE;
   char TempBuf[CONF_BUF_SIZE];

   if(fread(&Len,sizeof(Len),1,pCC->fp) != 1) {
      LOG_ERROR(("SendAudioFromFile(): fread failed, err=%d.\n",errno));
   }
   else if(Len > CONF_BUF_SIZE) {
      LOG_ERROR(("SendAudioFromFile(): Len > CONF_BUF_SIZE (%d).\n",Len));
   }
   else if(Len == 0) {
      D2PRINTF(("SendAudioFromFile(): playback complete.\n"));
   }
   else if(fread(TempBuf,Len,1,pCC->fp) != 1) {
      LOG_ERROR(("SendAudioFromFile(): fread failed, err=%d.\n",errno));
   }
   else {
      sendto(p->Socket,TempBuf,Len,0,&pCC->AdrRTP.s,sizeof(pCC->AdrRTP));
      Ret = TRUE;
   }

   if(!Ret) {
      PlayBackComplete(pCC);
   }
   
   return Ret;
}


int FilePlayBack(ClientInfo *p)
{
   int Delta;
   int bSendMore = TRUE;
   int Packets2Send = 0;

   ConfClient *pCC = (ConfClient *) p->p;

   if(pCC->TimeFirstTx.tv_sec != 0) {
      Delta = ((TimeNow.tv_sec - pCC->TimeFirstTx.tv_sec) * 1000) +
              ((TimeNow.tv_usec - pCC->TimeFirstTx.tv_usec) / 1000);

   // send one packet every 80 milliseconds
   // p->x = packets sent so far
      Packets2Send = (Delta / 80) - p->x;

   // Adjust next Timeout
      Delta = ((TimeNow.tv_sec - pCC->TimeLastTx.tv_sec) * 1000) +
              ((TimeNow.tv_usec - pCC->TimeLastTx.tv_usec) / 1000);

      pCC->Timeout -= ((Delta - 80)/2);
   }
   else {
   // First transmission
      pCC->TimeFirstTx = TimeNow;
      pCC->Timeout = 80;   // Assume a perfect delay at first
   }

   if(Packets2Send == 0) {
      Packets2Send++;
   }

   while(Packets2Send--) {
      if(!(bSendMore = SendAudioFromFile(p,pCC))) {
         break;
      }
      p->x++;
   }

   if(bSendMore) {
      GetTimeNow();
      pCC->TimeLastTx = TimeNow;
      SetTimeout(p,pCC->Timeout);
   }
   
   return FALSE;
}

int FileRecord(ClientInfo *p,ConfClient *pCC,EventType Event)
{
   ClientInfo *pPB;
   char *Filename;
   int Ret = TRUE;

   switch(Event) {
      case EVENT_INIT:
      // Open the file
         pCC->TimeFirstTx.tv_sec = 0;
         Filename = malloc(strlen(pCC->Callsign)+6);
         if(Filename != NULL) {
            strcpy(Filename,pCC->Callsign);
            strcat(Filename,".test");
            D2PRINTF(("FileRecord(): Opening file \"%s\".\n",Filename));
            pCC->fp = fopen(Filename,MODE_RDWR_BIN);
            free(Filename);
         }
         
         if(pCC->fp == NULL) {
         // file open failed
            LOG_ERROR(("FileRecord(): fopen failed, err=%d.\n",errno));
            Ret = FALSE;
         }
         break;

      case EVENT_RTP_RX:
         if(pCC->p == NULL && p->Buf[0] != ILINK_DATA_PACKET) {
            if(!pCC->bTalking) {
               D2PRINTF(("FileRecord(): First RTP data received.\n"));
               pCC->bTalking = TRUE;
            }
            pCC->LastAudioIn = TimeNow;
            if(fwrite(&p->Count,sizeof(p->Count),1,pCC->fp) != 1 ||
               fwrite(p->Buf,p->Count,1,pCC->fp) != 1) 
            {  // Write error
               LOG_ERROR(("FileRecord(): fwrite failed, err=%d.\n",errno));
               PlayBackComplete(pCC);
            }
         }
         break;

      case EVENT_RTP_TO:
      // timeout, recording is complete, play it back
         D2PRINTF(("FileRecord(): RTP timeout, rewinding file.\n"));
         p->Count = 0;
         if(fwrite(&p->Count,sizeof(p->Count),1,pCC->fp) != 1) {
         // Write error
            LOG_ERROR(("FileRecord(): fwrite failed, err=%d.\n",errno));
            PlayBackComplete(pCC);
         }
         else {
            fseek(pCC->fp,0,SEEK_SET);
         // Create a mainloop client to playback the file
            pPB = CreateNewClient();
            if(pPB == NULL) {
               LOG_ERROR(("FileRecord(): CreateNewClient failed.\n"));
               PlayBackComplete(pCC);
            }
            else {
            // Wait a second and then start playing it back
               pCC->p = pPB;
               pPB->p = pCC;
               pPB->HisAdr = pCC->AdrRTCP;
               pPB->Socket = p->Socket;
               pPB->State = FilePlayBack;
               SetTimeout(pPB,1000);
               avl_insert(ClientTree,pPB);
            }
         }
         break;

      case EVENT_RTCP_TO:
      // He's gone
         PlayBackComplete(pCC);
         break;

      default:
         break;
   }

   return Ret;
}

void CmdLevelTest(ClientInfo *p,ConfClient *pCC,const char *Arg)
{
   if(FileRecord(p,pCC,EVENT_INIT)) {
      pCC->bInConf = FALSE;
      pCC->State = FileRecord;
      BufPrintf(p,"Your next transmission will be recorded and played back.\r");
   }
}


// CmdLine
// 0 = command doesn't match
// 1 = command matches
// 2 = exact command match
int MatchCmd(const char *CmdLine, const char *Command)
{
   while(*CmdLine == ' ') {
      CmdLine++;
   }

   if(!*CmdLine) {
      return 0;
   }

   while(*CmdLine && *CmdLine != ' ' && *CmdLine != '\r' && *CmdLine != '\n') {
      if(tolower(*CmdLine) != tolower(*Command)) {
         return FALSE;
      }
      CmdLine++;
      Command++;
   }
   if(*Command == 0) {
   // Exact match
      return 2;
   }
   return 1;
}

struct COMMAND_TABLE {
   char *CmdString;
   char *HelpString;
   void (*CmdHandler)(ClientInfo *p,ConfClient *pCC,const char *Arg);
   int   Flags;
      #define  CMD_FLAG_NOLIST      1
      #define  CMD_FLAG_ADMIN       2
};

struct COMMAND_TABLE commandtable[] = {
   { "lurk", "Enter monitor mode", CmdLurk, 0 },
   { "delurk", "Leave monitor mode", CmdDelurk, 0 },
   { "lurkers", "", CmdLurkers, CMD_FLAG_NOLIST},
   { "uptime", "Display system uptime", CmdUpTime, 0},
   { "test", "Record and playback test transmission", CmdLevelTest, 0},
   { "version", "Display version info", CmdVersion, 0},
   { "help", "", CmdHelp, CMD_FLAG_NOLIST},
   { "?", "", CmdHelp, CMD_FLAG_NOLIST},
   { NULL}
};

void CmdHelp(ClientInfo *p,ConfClient *pCC,const char *Arg)
{
   struct COMMAND_TABLE *pTbl = commandtable;

   BufPrintf(p,"Commands:\r");
   
   pTbl = commandtable;
   while(pTbl->CmdString != NULL) {
      if(!(pTbl->Flags & CMD_FLAG_NOLIST) &&
         (!(pTbl->Flags & CMD_FLAG_ADMIN) || pCC->bAdmin))
      {
         BufPrintf(p,".%s:\r   %s\r",pTbl->CmdString,pTbl->HelpString);
      }
      pTbl++;
   }
}

// Process conference command
void ConferenceCmd(ClientInfo *p,ConfClient *pCC,char *CmdLine)
{
   int i;
   int Result;
   int InitialCount;
   int CurrentCmd = -1;
   int CmdsMatched = 0;
   const char *cp;

   p->Count = 0;  // init for BufPrintf
   BufPrintf(p,"%c" NDATA,ILINK_DATA_PACKET);
   InitialCount = p->Count;

   cp = CmdLine;
   while(*cp && *cp == ' ') {
      cp++;
   }

   if(*cp) {
      for(i = 0; commandtable[i].CmdString != NULL; i++) {
         if((Result = MatchCmd(CmdLine,commandtable[i].CmdString))) {
            CmdsMatched++;
            CurrentCmd = i;
            if(Result == 2) {
            // Exact match found
               CmdsMatched = 1;
               break;
            }
         }
      }

      if(CmdsMatched > 1) {
      // Ambiguous
         BufPrintf(p,"Error: Command is ambiguous.\r");
      }
      else if(CmdsMatched == 0) {
      // not found
      }
      else {
         while(*cp && *cp != ' ') {
            cp++;
         }
         while(*cp && *cp == ' ') {
            cp++;
         }
         commandtable[CurrentCmd].CmdHandler(p,pCC,cp);
      }
   }

   if(p->Count > InitialCount) {
   // send response
      sendto(p->Socket,p->Buf,p->Count+1,0,&pCC->AdrRTP.s,sizeof(pCC->AdrRTP));
   }
}

void DumpTree()
{
   struct avl_traverser avl_trans;
   ConfClient *pCC1;

   pCC1 = (ConfClient *) avl_t_first(&avl_trans,ConfTree);
   while(pCC1 != NULL) {
      LOG_NORM(("\tpCC=0x%x,SN=%d,IP=%s\n",pCC1,
                pCC1->SN,inet_ntoa(pCC1->AdrRTCP.i.sin_addr)));
      pCC1 = (ConfClient *) avl_t_next(&avl_trans);
   }
}

// Return number of milliseconds of time since p
int TimeLapse(struct timeval *p)
{
   return ((TimeNow.tv_sec - p->tv_sec) * 1000) + 
          ((TimeNow.tv_usec - p->tv_usec) / 1000);
}
