/* Copyright (C) 2002 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: configfile.c,v $
   Revision 1.1.1.1  2002/08/10 20:33:41  wb6ymh
   initial import

*/

#include "common.h"

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "configfile.h"

// Use a non-case sensitive compare function if one is available
#ifdef HAVE_STRICMP
   #define STRCMP stricmp
#else
#ifdef HAVE_STRCASECMP
   #define STRCMP strcasecmp
#else
   #define STRCMP strcmp
#endif
#endif

/*
   This routine is a general configuration file parser.
   The format of the configuration file is:
   variable=string;
   A structure describing the configuration variables and
   their type is passed to this routne.  since sscanf() is used to convert
   the value, all standard types are supported.  Use "%[^\n]" to input
   strings with embeded spaces.

*/

int is_comment(char *cp);

int read_config(char *filename,struct config_entry *config_vars)
{
   FILE *config;
   char line[121];
   char var_name[121];
   char var_value[121];
   char var_type_temp[4];
   int  var_temp;
   int i;
   struct config_entry *p;

   if((config = fopen(filename,"r")) == NULL){
      LOG_ERROR(("Fatal error opening configuration file, \"%s\".\n",filename));
      return CONFIG_OPEN_ERROR;
   }

   // clear the 'set' variable of all possible configuration variables

   for(i = 0; config_vars[i].var_name[0] != 0; i++){
      config_vars[i].set = 0;
   }

   while(fgets(line,120,config) != NULL){
      if(is_comment(line))    // ignore this line if it's a comment
         continue;

      if(sscanf(line," %[^= ] = %[^\n]",var_name,var_value) != 2){
         // line didn't convert
         line[strlen(line)-1] = 0;  // remove terminating <cr>
         LOG_WARN(("Error: Unable to parse \"%s\".\n       Line ignored.\n",
                   line));
         continue;
      }

      p = config_vars;
      for(i = 0; p->var_name[0] != 0; p++, i++){
         if(!STRCMP(var_name,p->var_name)){
            // found our guy
            if(p->set){
               line[strlen(line)-1] = 0;  // remove terminating <cr>
               // already set
               LOG_WARN(("Warning: Variable \"%s\" already set, ignoring "
                         "earlier value.\n",p->var_name));
            }
            if(p->var_type[0] == '%' && p->var_type[1] == 's' ) {
            // special case, string value
               *((char * *) p->var_ptr) = strdup(var_value);
               p->set = 1;
            }
            else if(p->var_type[0] == '%' && p->var_type[1] == 'b' ) {
            // special case, byte value
               var_type_temp[0] = '%';
               var_type_temp[1] = p->var_type[2];
               var_type_temp[2] = 0;
               if(sscanf(var_value,var_type_temp,&var_temp) != 1){
                  line[strlen(line)-1] = 0;  // remove terminating <cr>
                  LOG_WARN(("Error: Unable to convert \"%s\", line ignored.\n",
                            line));
               }
               else if(var_temp > 255 || var_temp < 0) {
                  line[strlen(line)-1] = 0;  // remove terminating <cr>
                  LOG_ERROR(("Error: Byte value out of range \"%s\", line ignored.\n"
                           ,line));
               }
               else {
                  *((char *) p->var_ptr) = (char) var_temp;
                  p->set = 1;
               }
            }
            else {
               if(sscanf(var_value,p->var_type,p->var_ptr) != 1){
                  line[strlen(line)-1] = 0;  // remove terminating <cr>
                  LOG_ERROR(("Error: Unable to convert \"%s\", Line ignored.\n",
                           line));
               }
               else{
                  p->set = 1;
               }
            }
            break;
         }
      }

      if(config_vars[i].var_name[0] == 0){
         LOG_ERROR(("Error: Unknown configuration variable \"%s\", ignored.\n",
                  var_name));
      }
   }

   fclose(config);

   // now check that all of the the required varables were set

   for(i = 0; config_vars[i].var_name[0] != 0; i++){
      if(config_vars[i].required && config_vars[i].set == 0){
         LOG_ERROR(("Error: Required configuration variable \"%s\" is not set.\n",
               config_vars[i].var_name));
         return CONFIG_REQUIRED_MISSING;
      }
   }
   return 0;
}

int is_comment(char *cp)
{
   while(*cp && isspace(*cp))
      cp++;
   if(*cp == ';' || *cp == '#' || *cp == 0 )
      return 1;
   return 0;
}

